@@ -613,7 +613,7 @@ static void sig_exec_input(PROCESS_REC *rec, const char *text)
 
 		str = g_strconcat(rec->target_nick ? "-nick " :
 				  rec->target_channel ? "-channel " : "",
-				  rec->target, " ", text, NULL);
+				  rec->target, " ", *text == '\0' ? " " : text, NULL);
 		signal_emit(rec->notice ? "command notice" : "command msg",
 			    3, str, server, item);
                 g_free(str);
